==> nono-0.4.4.tar.gz: nono-0.4.4/vm/test_busio.cpp <==
//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "header.h"
#include <stdint.h>
#include <vector>

// "device.h" の IODevice をリンクしてしまうと、芋づる式に依存オブジェクトが
// 増えてしまい収拾つかなくなるので、ダミークラスを用意する。
// なので、ここの実装は本物の IODevice のそれと同期しないといけない…。
class IODevice
{
 public:
	IODevice() {}
	virtual ~IODevice();

	virtual uint64 Read8(uint32 addr) { return 0xef; }
	virtual uint64 Read16(uint32 addr) {
		return (Read8(addr) << 8) | (Read8(addr + 1));
	}
	virtual uint64 Read32(uint32 addr) {
		return (Read16(addr) << 16) | (Read16(addr + 2));
	}
	virtual uint64 Write8(uint32 addr, uint32 data) { return 0; }
	virtual uint64 Write16(uint32 addr, uint32 data) {
		uint64 r;
		r  = Write8(addr,     data >> 8);
		r |= Write8(addr + 1, data & 0xff);
		return r;
	}
	virtual uint64 Write32(uint32 addr, uint32 data) {
		uint64 r;
		r  = Write16(addr,     data >> 16);
		r |= Write16(addr + 2, data & 0xffff);
		return r;
	}
	virtual uint64 Peek8(uint32 addr) { return 0xff; }
	virtual uint64 Poke8(uint32 addr, uint32 data) { return (uint64)-1; }
	virtual uint64 Poke(uint32 addr, uint32 data) { return (uint64)-1; }
};
IODevice::~IODevice() { }

#define SELFTEST
#include "busio.h"

class TestByteDev : public IODevice
{
 public:
	TestByteDev() {}
	virtual ~TestByteDev() override;

	static const uint32 NPORT = 4;
	uint32 reg[NPORT] {};
	uint64 Read(uint32 addr) {
		if ((int32)reg[addr] < 0)
			return (uint64)-1;
		return reg[addr];
	}
	uint64 Write(uint32 addr, uint32 data) {
		reg[addr] = data;
		return 0;
	}
	uint64 Peek(uint32 addr) {
		if ((int32)reg[addr] < 0)
			return (uint64)-1;
		return reg[addr];
	}
};
TestByteDev::~TestByteDev() { }

class TestWordDev : public IODevice
{
 public:
	TestWordDev() {}
	virtual ~TestWordDev() override;

	static const uint32 NPORT = 4;
	uint32 reg[NPORT] {};
	uint64 Read(uint32 addr) {
		if ((int32)reg[addr] < 0)
			return (uint64)-1;
		return reg[addr];
	}
	uint64 Write(uint32 addr, uint32 data) {
		reg[addr] = data;
		return 0;
	}
	uint64 Peek(uint32 addr) {
		if ((int32)reg[addr] < 0)
			return (uint64)-1;
		return reg[addr];
	}
};
TestWordDev::~TestWordDev() { }

// テストデータ
// expected[] の1エントリは1バイトを表す(WordDevの時も同様)。
#define E ((uint64)-1)
#define F (0xff)
static struct testdata {
	const char *name;
	std::vector<uint64> expected;
} testdata_b[] = {
	{ "BusIO_B", {
		1, 2, 3, 4, 1, 2, 3, 4,
	} },
	{ "BusIO_BB", {
		1, 1, 2, 2, 3, 3, 4, 4,  1, 1, 2, 2, 3, 3, 4, 4,
	} },
	{ "BusIO_BBBB", {
		1, 1, 1, 1, 2, 2, 2, 2,  3, 3, 3, 3, 4, 4, 4, 4,
		1, 1, 1, 1, 2, 2, 2, 2,  3, 3, 3, 3, 4, 4, 4, 4,
	} },
	{ "BusIO_BE", {
		1, E, 2, E, 3, E, 4, E,  1, E, 2, E, 3, E, 4, E,
	} },
	{ "BusIO_EB", {
		E, 1, E, 2, E, 3, E, 4,  E, 1, E, 2, E, 3, E, 4,
	} },
	{ "BusIO_FB", {
		F, 1, F, 2, F, 3, F, 4,  F, 1, F, 2, F, 3, F, 4,
	} },
	{ "BusIO_BFFF", {
		1, F, F, F, 2, F, F, F,  3, F, F, F, 4, F, F, F,
		1, F, F, F, 2, F, F, F,  3, F, F, F, 4, F, F, F,
	} },
}, testdata_w[] = {
	{ "BusIO_W", {
		1, 2, 3, 4, 5, 6, 7, 8,  1, 2, 3, 4, 5, 6, 7, 8,
	} },
	{ "BusIO_WW", {
		1, 2, 1, 2, 3, 4, 3, 4,  5, 6, 5, 6, 7, 8, 7, 8,
		1, 2, 1, 2, 3, 4, 3, 4,  5, 6, 5, 6, 7, 8, 7, 8,
	} },
	{ "BusIO_WF", {
		1, 2, F, F, 3, 4, F, F,  5, 6, F, F, 7, 8, F, F,
		1, 2, F, F, 3, 4, F, F,  5, 6, F, F, 7, 8, F, F,
	} },
};

void fail(const char *fmt, ...) __printflike(1, 2);
void
fail(const char *fmt, ...)
{
	va_list ap;

	printf(" FAIL: ");
	va_start(ap, fmt);
	vprintf(fmt, ap);
	va_end(ap);
}

template<typename T> void
do_test(struct testdata *test, T *dev)
{
	printf("%s\n", test->name);

	// デバイス側に読み出し元データをセット。
	// 元データは常にバイト単位で列挙してあるので、ターゲットデバイスの
	// サイズによってセット方法が変わる。
	if (dynamic_cast<TestByteDev*>(dev)) {
		for (int i = 0; i < dev->NPORT; i++) {
			dev->reg[i] = (i + 1);
		}
	} else {
		for (int i = 0, j = 0; i < dev->NPORT; i++) {
			uint32 data;
			data  = (++j) << 8;
			data |= (++j);
			dev->reg[i] = data;
		}
	}

	// Read8 のテスト
	for (int i = 0; i < test->expected.size(); i++) {
		uint64 act = dev->Read8(i);
		uint64 exp = test->expected[i];
		if (act != exp) {
			fail("%s::Read8(%d) expects %02" PRIx64 " but %02" PRIx64 "\n",
				test->name, i, exp, act);
		}
	}
	// Read16 のテスト
	for (int i = 0; i < test->expected.size(); i += 2) {
		uint64 act = dev->Read16(i);
		uint64 e0 = test->expected[i + 0];
		uint64 e1 = test->expected[i + 1];
		uint64 exp;
		if ((int64)e0 < 0 && (int64)e1 < 0) {
			// 両方バスエラーならバスエラー?
			exp = (int64)-1;
		} else {
			// そうでなければバスエラー側は 0xff ?
			exp = (((uint8)e0) << 8) | ((uint8)e1);
		}
		if (act != exp) {
			fail("%s::Read16(%d) expects %04" PRIx64 " but %04" PRIx64 "\n",
				test->name, i, exp, act);
		}
	}
	// Read32 のテスト
	for (int i = 0; i < test->expected.size(); i += 4) {
		uint64 act = dev->Read32(i);
		uint64 e0 = test->expected[i + 0];
		uint64 e1 = test->expected[i + 1];
		uint64 e2 = test->expected[i + 2];
		uint64 e3 = test->expected[i + 3];
		uint64 exp;
		if ((int64)e0 < 0 && (int64)e1 < 0 && (int64)e2 < 0 && (int64)e3 < 0) {
			// 全部バスエラーならバスエラー?
			exp = (int64)-1;
		} else {
			// そうでなければバスエラー部分だけ 0xff ?
			e0 = (uint8)e0 << 24;
			e1 = (uint8)e1 << 16;
			e2 = (uint8)e2 << 8;
			e3 = (uint8)e3;
			exp = (uint64)(uint32)(e0 | e1 | e2 | e3);
		}
		if (act != exp) {
			fail("%s::Read32(%d) expects %08" PRIx64 " but %08" PRIx64 "\n",
				test->name, i, exp, act);
		}
	}
	// Peek8 のテスト
	for (int i = 0; i < test->expected.size(); i++) {
		uint64 act = dev->Peek8(i);
		uint64 exp = test->expected[i];
		if (act != exp) {
			fail("%s::Peek8(%d) expects %02" PRIx64 " but %02" PRIx64 "\n",
				test->name, i, exp, act);
		}
	}

	// Write 系のテストはちょっと保留。
}

int
main()
{
	do_test<TestByteDev>(&testdata_b[0], new BusIO_B<TestByteDev>());
	do_test<TestByteDev>(&testdata_b[1], new BusIO_BB<TestByteDev>());
	do_test<TestByteDev>(&testdata_b[2], new BusIO_BBBB<TestByteDev>());
	do_test<TestByteDev>(&testdata_b[3], new BusIO_BE<TestByteDev>());
	do_test<TestByteDev>(&testdata_b[4], new BusIO_EB<TestByteDev>());
	do_test<TestByteDev>(&testdata_b[5], new BusIO_FB<TestByteDev>());
	do_test<TestByteDev>(&testdata_b[6], new BusIO_BFFF<TestByteDev>());

	do_test<TestWordDev>(&testdata_w[0], new BusIO_W<TestWordDev>());
	do_test<TestWordDev>(&testdata_w[1], new BusIO_WW<TestWordDev>());
	do_test<TestWordDev>(&testdata_w[2], new BusIO_WF<TestWordDev>());
	return 0;
}

// リンカを通すためのダミー
void
vmpanic(const char *funcname, const char *fmt, ...)
{
	va_list ap;

	printf("panic: %s: ", funcname);
	va_start(ap, fmt);
	vprintf(fmt, ap);
	va_end(ap);
	exit(1);
}

==> nono-0.6.3.tar.gz: nono-0.6.3/vm/test_busio.cpp <==
//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "header.h"
#include <stdint.h>
#include <vector>

// "device.h" の IODevice をリンクしてしまうと、芋づる式に依存オブジェクトが
// 増えてしまい収拾つかなくなるので、ダミークラスを用意する。
// なので、ここの実装は本物の IODevice のそれと同期しないといけない…。
class IODevice
{
 public:
	IODevice() {}
	virtual ~IODevice();

	virtual uint64 Read8(uint32 addr) { return 0xef; }
	virtual uint64 Read16(uint32 addr) {
		return (Read8(addr) << 8) | (Read8(addr + 1));
	}
	virtual uint64 Read32(uint32 addr) {
		return (Read16(addr) << 16) | (Read16(addr + 2));
	}
	virtual uint64 Write8(uint32 addr, uint32 data) { return 0; }
	virtual uint64 Write16(uint32 addr, uint32 data) {
		uint64 r;
		r  = Write8(addr,     data >> 8);
		r |= Write8(addr + 1, data & 0xff);
		return r;
	}
	virtual uint64 Write32(uint32 addr, uint32 data) {
		uint64 r;
		r  = Write16(addr,     data >> 16);
		r |= Write16(addr + 2, data & 0xffff);
		return r;
	}
	virtual uint64 Peek8(uint32 addr) { return 0xff; }
	virtual uint64 Poke8(uint32 addr, uint32 data) { return (uint64)-1; }
	virtual uint64 Poke(uint32 addr, uint32 data) { return (uint64)-1; }
};
IODevice::~IODevice() { }

#define SELFTEST
#include "busio.h"

class TestByteDev : public IODevice
{
 public:
	TestByteDev() {}
	virtual ~TestByteDev() override;

	static const uint32 NPORT = 4;
	uint32 reg[NPORT] {};
	uint64 Read(uint32 addr) {
		if ((int32)reg[addr] < 0)
			return (uint64)-1;
		return reg[addr];
	}
	uint64 Write(uint32 addr, uint32 data) {
		reg[addr] = data;
		return 0;
	}
	uint64 Peek(uint32 addr) {
		if ((int32)reg[addr] < 0)
			return (uint64)-1;
		return reg[addr];
	}
};
TestByteDev::~TestByteDev() { }

class TestWordDev : public IODevice
{
 public:
	TestWordDev() {}
	virtual ~TestWordDev() override;

	static const uint32 NPORT = 4;
	uint32 reg[NPORT] {};
	uint64 Read(uint32 addr) {
		if ((int32)reg[addr] < 0)
			return (uint64)-1;
		return reg[addr];
	}
	uint64 Write(uint32 addr, uint32 data) {
		reg[addr] = data;
		return 0;
	}
	uint64 Peek(uint32 addr) {
		if ((int32)reg[addr] < 0)
			return (uint64)-1;
		return reg[addr];
	}
};
TestWordDev::~TestWordDev() { }

// テストデータ
// expected[] の1エントリは1バイトを表す(WordDevの時も同様)。
#define E ((uint64)-1)
#define F (0xff)
static struct testdata {
	const char *name;
	std::vector<uint64> expected;
} testdata_b[] = {
	{ "BusIO_B", {
		1, 2, 3, 4, 1, 2, 3, 4,
	} },
	{ "BusIO_BB", {
		1, 1, 2, 2, 3, 3, 4, 4,  1, 1, 2, 2, 3, 3, 4, 4,
	} },
	{ "BusIO_BBBB", {
		1, 1, 1, 1, 2, 2, 2, 2,  3, 3, 3, 3, 4, 4, 4, 4,
		1, 1, 1, 1, 2, 2, 2, 2,  3, 3, 3, 3, 4, 4, 4, 4,
	} },
	{ "BusIO_BE", {
		1, E, 2, E, 3, E, 4, E,  1, E, 2, E, 3, E, 4, E,
	} },
	{ "BusIO_EB", {
		E, 1, E, 2, E, 3, E, 4,  E, 1, E, 2, E, 3, E, 4,
	} },
	{ "BusIO_FB", {
		F, 1, F, 2, F, 3, F, 4,  F, 1, F, 2, F, 3, F, 4,
	} },
	{ "BusIO_BFFF", {
		1, F, F, F, 2, F, F, F,  3, F, F, F, 4, F, F, F,
		1, F, F, F, 2, F, F, F,  3, F, F, F, 4, F, F, F,
	} },
}, testdata_w[] = {
	{ "BusIO_W", {
		1, 2, 3, 4, 5, 6, 7, 8,  1, 2, 3, 4, 5, 6, 7, 8,
	} },
	{ "BusIO_WW", {
		1, 2, 1, 2, 3, 4, 3, 4,  5, 6, 5, 6, 7, 8, 7, 8,
		1, 2, 1, 2, 3, 4, 3, 4,  5, 6, 5, 6, 7, 8, 7, 8,
	} },
	{ "BusIO_WF", {
		1, 2, F, F, 3, 4, F, F,  5, 6, F, F, 7, 8, F, F,
		1, 2, F, F, 3, 4, F, F,  5, 6, F, F, 7, 8, F, F,
	} },
};

void fail(const char *fmt, ...) __printflike(1, 2);
void
fail(const char *fmt, ...)
{
	va_list ap;

	printf(" FAIL: ");
	va_start(ap, fmt);
	vprintf(fmt, ap);
	va_end(ap);
}

template<typename T> void
do_test(struct testdata *test, T *dev)
{
	printf("%s\n", test->name);

	// デバイス側に読み出し元データをセット。
	// 元データは常にバイト単位で列挙してあるので、ターゲットデバイスの
	// サイズによってセット方法が変わる。
	if (dynamic_cast<TestByteDev*>(dev)) {
		for (int i = 0; i < dev->NPORT; i++) {
			dev->reg[i] = (i + 1);
		}
	} else {
		for (int i = 0, j = 0; i < dev->NPORT; i++) {
			uint32 data;
			data  = (++j) << 8;
			data |= (++j);
			dev->reg[i] = data;
		}
	}

	// Read8 のテスト
	for (int i = 0; i < test->expected.size(); i++) {
		uint64 act = dev->Read8(i);
		uint64 exp = test->expected[i];
		if (act != exp) {
			fail("%s::Read8(%d) expects %02" PRIx64 " but %02" PRIx64 "\n",
				test->name, i, exp, act);
		}
	}
	// Read16 のテスト
	for (int i = 0; i < test->expected.size(); i += 2) {
		uint64 act = dev->Read16(i);
		uint64 e0 = test->expected[i + 0];
		uint64 e1 = test->expected[i + 1];
		uint64 exp;
		if ((int64)e0 < 0 && (int64)e1 < 0) {
			// 両方バスエラーならバスエラー?
			exp = (int64)-1;
		} else {
			// そうでなければバスエラー側は 0xff ?
			exp = (((uint8)e0) << 8) | ((uint8)e1);
		}
		if (act != exp) {
			fail("%s::Read16(%d) expects %04" PRIx64 " but %04" PRIx64 "\n",
				test->name, i, exp, act);
		}
	}
	// Read32 のテスト
	for (int i = 0; i < test->expected.size(); i += 4) {
		uint64 act = dev->Read32(i);
		uint64 e0 = test->expected[i + 0];
		uint64 e1 = test->expected[i + 1];
		uint64 e2 = test->expected[i + 2];
		uint64 e3 = test->expected[i + 3];
		uint64 exp;
		if ((int64)e0 < 0 && (int64)e1 < 0 && (int64)e2 < 0 && (int64)e3 < 0) {
			// 全部バスエラーならバスエラー?
			exp = (int64)-1;
		} else {
			// そうでなければバスエラー部分だけ 0xff ?
			e0 = (uint8)e0 << 24;
			e1 = (uint8)e1 << 16;
			e2 = (uint8)e2 << 8;
			e3 = (uint8)e3;
			exp = (uint64)(uint32)(e0 | e1 | e2 | e3);
		}
		if (act != exp) {
			fail("%s::Read32(%d) expects %08" PRIx64 " but %08" PRIx64 "\n",
				test->name, i, exp, act);
		}
	}
	// Peek8 のテスト
	for (int i = 0; i < test->expected.size(); i++) {
		uint64 act = dev->Peek8(i);
		uint64 exp = test->expected[i];
		if (act != exp) {
			fail("%s::Peek8(%d) expects %02" PRIx64 " but %02" PRIx64 "\n",
				test->name, i, exp, act);
		}
	}

	// Write 系のテストはちょっと保留。
}

int
main()
{
	do_test<TestByteDev>(&testdata_b[0], new BusIO_B<TestByteDev>());
	do_test<TestByteDev>(&testdata_b[1], new BusIO_BB<TestByteDev>());
	do_test<TestByteDev>(&testdata_b[2], new BusIO_BBBB<TestByteDev>());
	do_test<TestByteDev>(&testdata_b[3], new BusIO_BE<TestByteDev>());
	do_test<TestByteDev>(&testdata_b[4], new BusIO_EB<TestByteDev>());
	do_test<TestByteDev>(&testdata_b[5], new BusIO_FB<TestByteDev>());
	do_test<TestByteDev>(&testdata_b[6], new BusIO_BFFF<TestByteDev>());

	do_test<TestWordDev>(&testdata_w[0], new BusIO_W<TestWordDev>());
	do_test<TestWordDev>(&testdata_w[1], new BusIO_WW<TestWordDev>());
	do_test<TestWordDev>(&testdata_w[2], new BusIO_WF<TestWordDev>());
	return 0;
}

// リンカを通すためのダミー
void
vmpanic_func(const char *funcname, const char *fmt, ...)
{
	va_list ap;

	printf("panic: %s: ", funcname);
	va_start(ap, fmt);
	vprintf(fmt, ap);
	va_end(ap);
	exit(1);
}
